@@ -7,16 +7,9 @@
 # (It may become useful if the test is moved to ./t subdirectory.)
 
 BEGIN { $| = 1; print "1..1\n"; }
-END {print "not ok 1\n" unless $loaded;}
+END {print " 1..........not ok\n" unless $loaded;}
 use DBI;
 use DDL::Oracle;
 $loaded = 1;
-print "Test 1: Checking that DBI and DDL::Oracle are installed...\n";
-print "ok 1\n";
-
-######################### End of black magic.
-
-# Insert your test code below (better if it prints "ok 13"
-# (correspondingly "not ok 13") depending on the success of chunk 13
-# of the test code):
+print " 1..........ok\n";
 
